@@ -1,6 +1,10 @@
 #include <mysql2_ext.h>
 #include <client.h>
 #include <errno.h>
+#ifndef _WIN32
+#include <sys/socket.h>
+#endif
+#include "wait_for_single_fd.h"
 
 VALUE cMysql2Client;
 extern VALUE mMysql2, cMysql2Error;
@@ -9,7 +13,7 @@ static VALUE sym_id, sym_version, sym_async, sym_symbolize_keys, sym_as, sym_arr
 static ID intern_merge, intern_error_number_eql, intern_sql_state_eql;
 
 #define REQUIRE_OPEN_DB(wrapper) \
-  if(wrapper->closed) { \
+  if(!wrapper->reconnect_enabled && wrapper->closed) { \
     rb_raise(cMysql2Error, "closed MySQL connection"); \
   }
 
@@ -42,6 +46,7 @@ struct nogvl_connect_args {
 struct nogvl_send_query_args {
   MYSQL *mysql;
   VALUE sql;
+  mysql_client_wrapper *wrapper;
 };
 
 /*
@@ -125,7 +130,7 @@ static VALUE nogvl_close(void *ptr) {
   wrapper = ptr;
   if (!wrapper->closed) {
     wrapper->closed = 1;
-
+    wrapper->active = 0;
     /*
      * we'll send a QUIT message to the server, but that message is more of a
      * formality than a hard requirement since the socket is getting shutdown
@@ -139,14 +144,10 @@ static VALUE nogvl_close(void *ptr) {
     flags = fcntl(wrapper->client->net.fd, F_GETFL);
     if (flags > 0 && !(flags & O_NONBLOCK))
       fcntl(wrapper->client->net.fd, F_SETFL, flags | O_NONBLOCK);
-#else
-    u_long iMode;
-    iMode = 1;
-    ioctlsocket(wrapper->client->net.fd, FIONBIO, &iMode);
 #endif
 
     mysql_close(wrapper->client);
-    xfree(wrapper->client);
+    free(wrapper->client);
   }
 
   return Qnil;
@@ -157,7 +158,7 @@ static void rb_mysql_client_free(void * ptr) {
 
   nogvl_close(wrapper);
 
-  xfree(ptr);
+  free(ptr);
 }
 
 static VALUE allocate(VALUE klass) {
@@ -166,11 +167,37 @@ static VALUE allocate(VALUE klass) {
   obj = Data_Make_Struct(klass, mysql_client_wrapper, rb_mysql_client_mark, rb_mysql_client_free, wrapper);
   wrapper->encoding = Qnil;
   wrapper->active = 0;
+  wrapper->reconnect_enabled = 0;
   wrapper->closed = 1;
-  wrapper->client = (MYSQL*)xmalloc(sizeof(MYSQL));
+  wrapper->client = (MYSQL*)malloc(sizeof(MYSQL));
   return obj;
 }
 
+static VALUE rb_mysql_client_escape(RB_MYSQL_UNUSED VALUE klass, VALUE str) {
+  unsigned char *newStr;
+  VALUE rb_str;
+  unsigned long newLen, oldLen;
+
+  Check_Type(str, T_STRING);
+
+  oldLen = RSTRING_LEN(str);
+  newStr = malloc(oldLen*2+1);
+
+  newLen = mysql_escape_string((char *)newStr, StringValuePtr(str), oldLen);
+  if (newLen == oldLen) {
+    // no need to return a new ruby string if nothing changed
+    free(newStr);
+    return str;
+  } else {
+    rb_str = rb_str_new((const char*)newStr, newLen);
+#ifdef HAVE_RUBY_ENCODING_H
+    rb_enc_copy(rb_str, str);
+#endif
+    free(newStr);
+    return rb_str;
+  }
+}
+
 static VALUE rb_connect(VALUE self, VALUE user, VALUE pass, VALUE host, VALUE port, VALUE database, VALUE socket, VALUE flags) {
   struct nogvl_connect_args args;
   GET_CLIENT(self);
@@ -224,6 +251,17 @@ static VALUE nogvl_send_query(void *ptr) {
   return rv == 0 ? Qtrue : Qfalse;
 }
 
+static VALUE do_send_query(void *args) {
+  struct nogvl_send_query_args *query_args = args;
+  mysql_client_wrapper *wrapper = query_args->wrapper;
+  if (rb_thread_blocking_region(nogvl_send_query, args, RUBY_UBF_IO, 0) == Qfalse) {
+    // an error occurred, we're not active anymore
+    MARK_CONN_INACTIVE(self);
+    return rb_raise_mysql2_error(wrapper);
+  }
+  return Qnil;
+}
+
 /*
  * even though we did rb_thread_select before calling this, a large
  * response can overflow the socket buffers and cause us to eventually
@@ -238,8 +276,17 @@ static VALUE nogvl_read_query_result(void *ptr) {
 
 /* mysql_store_result may (unlikely) read rows off the socket */
 static VALUE nogvl_store_result(void *ptr) {
-  MYSQL * client = ptr;
-  return (VALUE)mysql_store_result(client);
+  mysql_client_wrapper *wrapper;
+  MYSQL_RES *result;
+
+  wrapper = (mysql_client_wrapper *)ptr;
+  result = mysql_store_result(wrapper->client);
+
+  // once our result is stored off, this connection is
+  // ready for another command to be issued
+  wrapper->active = 0;
+
+  return (VALUE)result;
 }
 
 static VALUE rb_mysql_client_async_result(VALUE self) {
@@ -250,6 +297,10 @@ static VALUE rb_mysql_client_async_result(VALUE self) {
 #endif
   GET_CLIENT(self);
 
+  // if we're not waiting on a result, do nothing
+  if (!wrapper->active)
+    return Qnil;
+
   REQUIRE_OPEN_DB(wrapper);
   if (rb_thread_blocking_region(nogvl_read_query_result, wrapper->client, RUBY_UBF_IO, 0) == Qfalse) {
     // an error occurred, mark this connection inactive
@@ -257,10 +308,7 @@ static VALUE rb_mysql_client_async_result(VALUE self) {
     return rb_raise_mysql2_error(wrapper);
   }
 
-  result = (MYSQL_RES *)rb_thread_blocking_region(nogvl_store_result, wrapper->client, RUBY_UBF_IO, 0);
-
-  // we have our result, mark this connection inactive
-  MARK_CONN_INACTIVE(self);
+  result = (MYSQL_RES *)rb_thread_blocking_region(nogvl_store_result, wrapper, RUBY_UBF_IO, 0);
 
   if (result == NULL) {
     if (mysql_field_count(wrapper->client) != 0) {
@@ -280,31 +328,109 @@ static VALUE rb_mysql_client_async_result(VALUE self) {
   return resultObj;
 }
 
-static VALUE rb_mysql_client_query(int argc, VALUE * argv, VALUE self) {
-  struct nogvl_send_query_args args;
-  fd_set fdset;
-  int fd, retval;
-  int async = 0;
-  VALUE opts, defaults, read_timeout;
-#ifdef HAVE_RUBY_ENCODING_H
-  rb_encoding *conn_enc;
-#endif
+#ifndef _WIN32
+struct async_query_args {
+  int fd;
+  VALUE self;
+};
+
+static VALUE disconnect_and_raise(VALUE self, VALUE error) {
+  GET_CLIENT(self);
+
+  wrapper->closed = 1;
+  wrapper->active = 0;
+
+  // manually close the socket for read/write
+  // this feels dirty, but is there another way?
+  shutdown(wrapper->client->net.fd, 2);
+
+  rb_exc_raise(error);
+
+  return Qnil;
+}
+
+static VALUE do_query(void *args) {
+  struct async_query_args *async_args;
   struct timeval tv;
   struct timeval* tvp;
   long int sec;
-  VALUE result;
+  int retval;
+  VALUE read_timeout;
+
+  async_args = (struct async_query_args *)args;
+  read_timeout = rb_iv_get(async_args->self, "@read_timeout");
+
+  tvp = NULL;
+  if (!NIL_P(read_timeout)) {
+    Check_Type(read_timeout, T_FIXNUM);
+    tvp = &tv;
+    sec = FIX2INT(read_timeout);
+    // TODO: support partial seconds?
+    // also, this check is here for sanity, we also check up in Ruby
+    if (sec >= 0) {
+      tvp->tv_sec = sec;
+    } else {
+      rb_raise(cMysql2Error, "read_timeout must be a positive integer, you passed %ld", sec);
+    }
+    tvp->tv_usec = 0;
+  }
+
+  for(;;) {
+    retval = rb_wait_for_single_fd(async_args->fd, RB_WAITFD_IN, tvp);
+
+    if (retval == 0) {
+      rb_raise(cMysql2Error, "Timeout waiting for a response from the last query. (waited %d seconds)", FIX2INT(read_timeout));
+    }
+
+    if (retval < 0) {
+      rb_sys_fail(0);
+    }
+
+    if (retval > 0) {
+      break;
+    }
+  }
+
+  return Qnil;
+}
+#else
+static VALUE finish_and_mark_inactive(void *args) {
+  VALUE self;
+  MYSQL_RES *result;
+
+  self = (VALUE)args;
+
+  GET_CLIENT(self);
+
+  if (wrapper->active) {
+    // if we got here, the result hasn't been read off the wire yet
+    // so lets do that and then throw it away because we have no way
+    // of getting it back up to the caller from here
+    result = (MYSQL_RES *)rb_thread_blocking_region(nogvl_store_result, wrapper, RUBY_UBF_IO, 0);
+    mysql_free_result(result);
+
+    wrapper->active = 0;
+  }
+
+  return Qnil;
+}
+#endif
+
+static VALUE rb_mysql_client_query(int argc, VALUE * argv, VALUE self) {
+#ifndef _WIN32
+  struct async_query_args async_args;
+#endif
+  struct nogvl_send_query_args args;
+  int async = 0;
+  VALUE opts, defaults;
+#ifdef HAVE_RUBY_ENCODING_H
+  rb_encoding *conn_enc;
+#endif
   GET_CLIENT(self);
 
   REQUIRE_OPEN_DB(wrapper);
   args.mysql = wrapper->client;
 
-  // see if this connection is still waiting on a result from a previous query
-  if (wrapper->active == 0) {
-    // mark this connection active
-    wrapper->active = 1;
-  } else {
-    rb_raise(cMysql2Error, "This connection is still waiting for a result, try again once you have the result");
-  }
 
   defaults = rb_iv_get(self, "@query_options");
   if (rb_scan_args(argc, argv, "11", &args.sql, &opts) == 2) {
@@ -325,79 +451,38 @@ static VALUE rb_mysql_client_query(int argc, VALUE * argv, VALUE self) {
   args.sql = rb_str_export_to_enc(args.sql, conn_enc);
 #endif
 
-  if (rb_thread_blocking_region(nogvl_send_query, &args, RUBY_UBF_IO, 0) == Qfalse) {
-    // an error occurred, we're not active anymore
-    MARK_CONN_INACTIVE(self);
-    return rb_raise_mysql2_error(wrapper);
+  // see if this connection is still waiting on a result from a previous query
+  if (wrapper->active == 0) {
+    // mark this connection active
+    wrapper->active = 1;
+  } else {
+    rb_raise(cMysql2Error, "This connection is still waiting for a result, try again once you have the result");
   }
 
-  read_timeout = rb_iv_get(self, "@read_timeout");
+  args.wrapper = wrapper;
 
-  tvp = NULL;
-  if (!NIL_P(read_timeout)) {
-    Check_Type(read_timeout, T_FIXNUM);
-    tvp = &tv;
-    sec = FIX2INT(read_timeout);
-    // TODO: support partial seconds?
-    // also, this check is here for sanity, we also check up in Ruby
-    if (sec >= 0) {
-      tvp->tv_sec = sec;
-    } else {
-      rb_raise(cMysql2Error, "read_timeout must be a positive integer, you passed %ld", sec);
-    }
-    tvp->tv_usec = 0;
-  }
+#ifndef _WIN32
+  rb_rescue2(do_send_query, (VALUE)&args, disconnect_and_raise, self, rb_eException, (VALUE)0);
 
   if (!async) {
-    // the below code is largely from do_mysql
-    // http://github.com/datamapper/do
-    fd = wrapper->client->net.fd;
-    for(;;) {
-      int fd_set_fd = fd;
+    async_args.fd = wrapper->client->net.fd;
+    async_args.self = self;
 
-#ifdef _WIN32
-      WSAPROTOCOL_INFO wsa_pi;
-      // dupicate the SOCKET from libmysql
-      int r = WSADuplicateSocket(fd, GetCurrentProcessId(), &wsa_pi);
-      SOCKET s = WSASocket(wsa_pi.iAddressFamily, wsa_pi.iSocketType, wsa_pi.iProtocol, &wsa_pi, 0, 0);
-      // create the CRT fd so ruby can get back to the SOCKET
-      fd_set_fd = _open_osfhandle(s, O_RDWR|O_BINARY);
-#endif
+    rb_rescue2(do_query, (VALUE)&async_args, disconnect_and_raise, self, rb_eException, (VALUE)0);
 
-      FD_ZERO(&fdset);
-      FD_SET(fd_set_fd, &fdset);
-
-      retval = rb_thread_select(fd_set_fd + 1, &fdset, NULL, NULL, tvp);
-
-#ifdef _WIN32
-      // cleanup the CRT fd
-      _close(fd_set_fd);
-      // cleanup the duplicated SOCKET
-      closesocket(s);
-#endif
-
-      if (retval == 0) {
-        rb_raise(cMysql2Error, "Timeout waiting for a response from the last query. (waited %d seconds)", FIX2INT(read_timeout));
-      }
-
-      if (retval < 0) {
-        rb_sys_fail(0);
-      }
-
-      if (retval > 0) {
-        break;
-      }
-    }
-
-    result = rb_mysql_client_async_result(self);
-
-    return result;
+    return rb_mysql_client_async_result(self);
   } else {
     return Qnil;
   }
+#else
+  do_send_query(&args);
+
+  // this will just block until the result is ready
+  return rb_ensure(rb_mysql_client_async_result, self, finish_and_mark_inactive, self);
+#endif
 }
 
-static VALUE rb_mysql_client_escape(VALUE self, VALUE str) {
+static VALUE rb_mysql_client_real_escape(VALUE self, VALUE str) {
   unsigned char *newStr;
   VALUE rb_str;
   unsigned long newLen, oldLen;
@@ -417,12 +502,12 @@ static VALUE rb_mysql_client_escape(VALUE self, VALUE str) {
 #endif
 
   oldLen = RSTRING_LEN(str);
-  newStr = xmalloc(oldLen*2+1);
+  newStr = malloc(oldLen*2+1);
 
   newLen = mysql_real_escape_string(wrapper->client, (char *)newStr, StringValuePtr(str), oldLen);
   if (newLen == oldLen) {
     // no need to return a new ruby string if nothing changed
-    xfree(newStr);
+    free(newStr);
     return str;
   } else {
     rb_str = rb_str_new((const char*)newStr, newLen);
@@ -432,7 +517,7 @@ static VALUE rb_mysql_client_escape(VALUE self, VALUE str) {
       rb_str = rb_str_export_to_enc(rb_str, default_internal_enc);
     }
 #endif
-    xfree(newStr);
+    free(newStr);
     return rb_str;
   }
 }
@@ -492,8 +577,13 @@ static VALUE rb_mysql_client_server_info(VALUE self) {
 
 static VALUE rb_mysql_client_socket(VALUE self) {
   GET_CLIENT(self);
+#ifndef _WIN32
   REQUIRE_OPEN_DB(wrapper);
-  return INT2NUM(wrapper->client->net.fd);
+  int fd_set_fd = wrapper->client->net.fd;
+  return INT2NUM(fd_set_fd);
+#else
+  rb_raise(cMysql2Error, "Raw access to the mysql file descriptor isn't supported on Windows");
+#endif
 }
 
 static VALUE rb_mysql_client_last_id(VALUE self) {
@@ -523,8 +613,7 @@ static VALUE rb_mysql_client_thread_id(VALUE self) {
   return ULL2NUM(retVal);
 }
 
-static VALUE nogvl_ping(void *ptr)
-{
+static VALUE nogvl_ping(void *ptr) {
   MYSQL *client = ptr;
 
   return mysql_ping(client) == 0 ? Qtrue : Qfalse;
@@ -533,7 +622,11 @@ static VALUE nogvl_ping(void *ptr)
 static VALUE rb_mysql_client_ping(VALUE self) {
   GET_CLIENT(self);
 
-  return rb_thread_blocking_region(nogvl_ping, wrapper->client, RUBY_UBF_IO, 0);
+  if (wrapper->closed) {
+    return Qfalse;
+  } else {
+    return rb_thread_blocking_region(nogvl_ping, wrapper->client, RUBY_UBF_IO, 0);
+  }
 }
 
 #ifdef HAVE_RUBY_ENCODING_H
@@ -550,6 +643,7 @@ static VALUE set_reconnect(VALUE self, VALUE value) {
   if(!NIL_P(value)) {
     reconnect = value == Qfalse ? 0 : 1;
 
+    wrapper->reconnect_enabled = reconnect;
     /* set default reconnect behavior */
     if (mysql_options(wrapper->client, MYSQL_OPT_RECONNECT, &reconnect)) {
       /* TODO: warning - unable to set reconnect behavior */
@@ -654,9 +748,11 @@ void init_mysql2_client() {
 
   rb_define_alloc_func(cMysql2Client, allocate);
 
+  rb_define_singleton_method(cMysql2Client, "escape", rb_mysql_client_escape, 1);
+
   rb_define_method(cMysql2Client, "close", rb_mysql_client_close, 0);
   rb_define_method(cMysql2Client, "query", rb_mysql_client_query, -1);
-  rb_define_method(cMysql2Client, "escape", rb_mysql_client_escape, 1);
+  rb_define_method(cMysql2Client, "escape", rb_mysql_client_real_escape, 1);
   rb_define_method(cMysql2Client, "info", rb_mysql_client_info, 0);
   rb_define_method(cMysql2Client, "server_info", rb_mysql_client_server_info, 0);
   rb_define_method(cMysql2Client, "socket", rb_mysql_client_socket, 0);
